use std::iter::FusedIterator;

#[cfg(feature = "derive")]
pub use stable_step_derive::Step;

pub trait Step: PartialOrd {
    const MIN: Self;
    const MAX: Self;

    fn next(&self) -> Option<Self>
    where
        Self: Sized;

    fn prev(&self) -> Option<Self>
    where
        Self: Sized;
}

pub trait StepExt: Step {
    fn iter() -> RangeIter<Self>
    where
        Self: Sized,
    {
        range(None, None)
    }

    fn iter_to(self, to: Self) -> RangeIter<Self>
    where
        Self: Sized,
    {
        range(Some(self), Some(to))
    }

    fn iter_to_inclusive(self, to: Self) -> RangeIter<Self>
    where
        Self: Sized,
    {
        range_inclusive(Some(self), Some(to))
    }
}

impl<S> StepExt for S where S: Step {}

pub struct RangeIter<S> {
    lower: Option<S>,
    upper: Option<S>,
}

pub fn range_inclusive<S: Step>(from: Option<S>, to: Option<S>) -> RangeIter<S> {
    let from = match from {
        Some(lower) => lower,
        None => S::MIN,
    };

    let to = match to {
        Some(upper) => upper,
        None => S::MAX,
    };

    RangeIter {
        lower: Some(from),
        upper: Some(to),
    }
}

pub fn range<S: Step>(from: Option<S>, to: Option<S>) -> RangeIter<S> {
    let to = to.and_then(|to| to.prev());

    range_inclusive(from, to)
}

impl<S> Iterator for RangeIter<S>
where
    S: Step,
{
    type Item = S;

    fn next(&mut self) -> Option<Self::Item> {
        let current = self.lower.take()?;

        let upper = self.upper.as_ref()?;

        if &current > upper {
            return None;
        }

        self.lower = current.next();

        Some(current)
    }
}

impl<S> DoubleEndedIterator for RangeIter<S>
where
    S: Step,
{
    fn next_back(&mut self) -> Option<Self::Item> {
        let current = self.upper.take()?;

        let lower = self.lower.as_ref()?;

        if &current < lower {
            return None;
        }

        self.upper = current.prev();

        Some(current)
    }
}

impl<S> FusedIterator for RangeIter<S> where S: Step {}

#[cfg(test)]
mod tests {
    use super::{range, range_inclusive, Step};

    #[derive(Debug, PartialEq, Eq, PartialOrd, Ord)]
    enum Test {
        One,
        Two,
        Three,
        Four,
    }

    impl Step for Test {
        const MIN: Self = Self::One;
        const MAX: Self = Self::Four;

        fn next(&self) -> Option<Self>
        where
            Self: Sized,
        {
            match self {
                Self::One => Some(Self::Two),
                Self::Two => Some(Self::Three),
                Self::Three => Some(Self::Four),
                Self::Four => None,
            }
        }

        fn prev(&self) -> Option<Self>
        where
            Self: Sized,
        {
            match self {
                Self::One => None,
                Self::Two => Some(Self::One),
                Self::Three => Some(Self::Two),
                Self::Four => Some(Self::Three),
            }
        }
    }

    #[test]
    fn range_none_none() {
        let v: Vec<Test> = range(None, None).collect();

        assert_eq!(v, vec![Test::One, Test::Two, Test::Three, Test::Four]);
    }

    #[test]
    fn range_reverse_none_none() {
        let v: Vec<Test> = range(None, None).rev().collect();

        assert_eq!(v, vec![Test::Four, Test::Three, Test::Two, Test::One]);
    }

    #[test]
    fn range_upper() {
        let v: Vec<Test> = range(None, Some(Test::Four)).collect();

        assert_eq!(v, vec![Test::One, Test::Two, Test::Three]);
    }

    #[test]
    fn range_reverse_upper() {
        let v: Vec<Test> = range(None, Some(Test::Four)).rev().collect();

        assert_eq!(v, vec![Test::Three, Test::Two, Test::One]);
    }

    #[test]
    fn range_lower() {
        let v: Vec<Test> = range(Some(Test::Two), None).collect();

        assert_eq!(v, vec![Test::Two, Test::Three, Test::Four]);
    }

    #[test]
    fn range_reverse_lower() {
        let v: Vec<Test> = range(Some(Test::Two), None).rev().collect();

        assert_eq!(v, vec![Test::Four, Test::Three, Test::Two]);
    }

    #[test]
    fn range_upper_lower() {
        let v: Vec<Test> = range(Some(Test::Two), Some(Test::Four)).collect();

        assert_eq!(v, vec![Test::Two, Test::Three]);
    }

    #[test]
    fn range_reverse_upper_lower() {
        let v: Vec<Test> = range(Some(Test::Two), Some(Test::Four)).rev().collect();

        assert_eq!(v, vec![Test::Three, Test::Two]);
    }

    #[test]
    fn range_inclusive_none_none() {
        let v: Vec<Test> = range_inclusive(None, None).collect();

        assert_eq!(v, vec![Test::One, Test::Two, Test::Three, Test::Four]);
    }

    #[test]
    fn range_reverse_inclusive_none_none() {
        let v: Vec<Test> = range_inclusive(None, None).rev().collect();

        assert_eq!(v, vec![Test::Four, Test::Three, Test::Two, Test::One]);
    }

    #[test]
    fn range_inclusive_upper() {
        let v: Vec<Test> = range_inclusive(None, Some(Test::Three)).collect();

        assert_eq!(v, vec![Test::One, Test::Two, Test::Three]);
    }

    #[test]
    fn range_reverse_inclusive_upper() {
        let v: Vec<Test> = range_inclusive(None, Some(Test::Three)).rev().collect();

        assert_eq!(v, vec![Test::Three, Test::Two, Test::One]);
    }

    #[test]
    fn range_inclusive_lower() {
        let v: Vec<Test> = range_inclusive(Some(Test::Two), None).collect();

        assert_eq!(v, vec![Test::Two, Test::Three, Test::Four]);
    }

    #[test]
    fn range_reverse_inclusive_lower() {
        let v: Vec<Test> = range_inclusive(Some(Test::Two), None).rev().collect();

        assert_eq!(v, vec![Test::Four, Test::Three, Test::Two]);
    }

    #[test]
    fn range_inclusive_upper_lower() {
        let v: Vec<Test> = range_inclusive(Some(Test::Two), Some(Test::Three)).collect();

        assert_eq!(v, vec![Test::Two, Test::Three]);
    }

    #[test]
    fn range_reverse_inclusive_upper_lower() {
        let v: Vec<Test> = range_inclusive(Some(Test::Two), Some(Test::Three))
            .rev()
            .collect();

        assert_eq!(v, vec![Test::Three, Test::Two]);
    }
}
